﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Boats
{
    internal abstract class AbstractMap
    {
        private IDrawingObject _drawingObject = null;
        protected int[,] _map = null;
        protected int _width;
        protected int _height;
        protected float _size_x;
        protected float _size_y;
        protected readonly Random _random = new();
        protected readonly int _freeWater = 0;
        protected readonly int _barrier = 1;
        /// <summary>
        /// Функция инициализирует карту,
        /// устанавливает объект и отрисовывает карту с объектом
        /// </summary>
        /// <param name="width"></param>
        /// <param name="height"></param>
        /// <param name="drawingObject"></param>
        /// <returns>Bitmap карты с объектом</returns>
        public Bitmap CreateMap(int width, int height, IDrawingObject drawingObject)
        {
            _width = width;
            _height = height;
            _drawingObject = drawingObject;
            GenerateMap();
            while (!SetObjectOnMap())
            {
                GenerateMap();
            }
            return DrawMapWithObject();
        }
        /// <summary>
        /// Функция для передвижения объекта по карте
        /// </summary>
        /// <param name="direction"></param>
        /// <returns>Bitmap карты с объектом</returns>
        public Bitmap MoveObject(Direction direction)
        {
            // Проверка, что объект может переместится в требуемом направлении
            // Получаем текщую позицию объекта
            var objectPos = _drawingObject.GetCurrentPosition();
            float currentX = objectPos.Left;
            float currentY = objectPos.Top;

            // В зависимости от направления уставналиваем dx, dy
            float dx = 0;
            float dy = 0;
            switch (direction)
            {
                case Direction.None:
                    break;
                case Direction.Up:
                    {
                        dy = -_drawingObject.Step;
                    }
                    break;
                case Direction.Down:
                    {
                        dy = _drawingObject.Step;
                    }
                    break;
                case Direction.Left:
                    {
                        dx = -_drawingObject.Step;
                    }
                    break;
                case Direction.Right:
                    {
                        dx = _drawingObject.Step;
                    }
                    break;
                default:
                    break;
            }
            // ЕСли нет коллизии, то перемещаем объект
            if (!CheckCollision(currentX + dx, currentY + dy))
            {
                _drawingObject.MoveObject(direction);
            }
            return DrawMapWithObject();
        }
        /// <summary>
        /// Функция пытается поместить объект на карту
        /// </summary>
        /// <returns>Если удачно возвращает true, иначе - false</returns>
        private bool SetObjectOnMap()
        {
            if (_drawingObject == null || _map == null)
            {
                return false;
            }
            // Генерируем новые координаты объекта
            int x = _random.Next(100, 200);
            int y = _random.Next(100, 200);
            _drawingObject.SetObject(x, y, _width, _height);

            // Проверка, что объект не "накладывается" на закрытые участки
            return !CheckCollision(x, y);
        }
        /// <summary>
        /// Функция для проверки коллизии объекта с препятствиями на карте.
        /// </summary>
        /// <param name="x">Координата x объекта</param>
        /// <param name="y">Координата y объекта</param>
        /// <returns>Возвращает true если есть коллизия и false - если ее нет</returns>
        protected bool CheckCollision(float x, float y)
        {
            // Получаем ширину и высоту отображаемого объекта
            var objectPos = _drawingObject.GetCurrentPosition();
            float objectWidth = objectPos.Right - objectPos.Left;
            float objectHeight = objectPos.Bottom - objectPos.Top;

            // Теперь узнаем сколько клеток в ширину и высоту объект занимает на карте
            int objectCellsCountX = (int)Math.Ceiling(objectWidth / _size_x);
            int objectCellsCountY = (int)Math.Ceiling(objectHeight / _size_y);

            // Получим координаты объекта в сетке карты
            int objectMapX = (int)Math.Floor((float)x / _size_x);
            int objectMapY = (int)Math.Floor((float)y / _size_y);

            // В цикле проверяем все клетки карты на коллизию с объектом
            int dy = 0;
            int mapCellsX = _map.GetLength(1);
            int mapCellsY = _map.GetLength(0);
            int mapState = _freeWater;

            while (objectMapY + dy < mapCellsY && dy <= objectCellsCountY)
            {
                int dx = 0;
                while (objectMapX + dx < mapCellsX && dx <= objectCellsCountX)
                {
                    mapState = _map[objectMapX + dx, objectMapY + dy];
                    if (mapState == _barrier)
                    {
                        return true;
                    }
                    dx++;
                }
                dy++;
            }
            return false;
        }
        /// <summary>
        /// Функция для отрисовки карты с объектом
        /// </summary>
        /// <returns></returns>
        private Bitmap DrawMapWithObject()
        {
            Bitmap bmp = new(_width, _height);
            if (_drawingObject == null || _map == null)
            {
                return bmp;
            }
            Graphics g = Graphics.FromImage(bmp);

            for (int i = 0; i < _map.GetLength(0); ++i)
            {
                for (int j = 0; j < _map.GetLength(1); ++j)
                {
                    if (_map[i, j] == _freeWater)
                    {
                        DrawWaterPart(g, i, j);
                    }
                    else if (_map[i, j] == _barrier)
                    {
                        DrawBarrierPart(g, i, j);
                    }
                }
            }
            _drawingObject.DrawingObject(g);
            return bmp;
        }
        /// <summary>
        /// Метод для генерации карты
        /// </summary>
        protected abstract void GenerateMap();
        /// <summary>
        /// Метод для отрисовки свободного участка на экране
        /// </summary>
        /// <param name="g"></param>
        /// <param name="i"></param>
        /// <param name="j"></param>
        protected abstract void DrawWaterPart(Graphics g, int i, int j);
        /// <summary>
        /// Метод для отрисовки закрытого участка на экране
        /// </summary>
        /// <param name="g"></param>
        /// <param name="i"></param>
        /// <param name="j"></param>
        protected abstract void DrawBarrierPart(Graphics g, int i, int j);
    }
}
