﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using static System.Net.Mime.MediaTypeNames;

namespace Boats
{
    /// <summary>
    /// Класс для хранения коллекции карт
    /// </summary>
    internal class MapsCollection
    {
        /// <summary>
        /// Словарь (хранилище) с картами
        /// </summary>
        readonly Dictionary<string, MapWithSetBoatsGeneric<IDrawingObject, AbstractMap>> _mapStorages;
        /// <summary>
        /// Возвращение списка названий карт
        /// </summary>
        public List<string> Keys => _mapStorages.Keys.ToList();
        /// <summary>
        /// Ширина окна отрисовки
        /// </summary>
        private readonly int _pictureWidth;
        /// <summary>
        /// Высота окна отрисовки
        /// </summary>
        private readonly int _pictureHeight;
        /// <summary>
        /// Разделитель для записи информации по элементу словаря в файл
        /// </summary>
        private readonly char separatorDict = '|';
        /// <summary>
        /// Разделитель для записи коллекции данных в файл
        /// </summary>
        private readonly char separatorData = ';';
        /// <summary>
        /// Конструктор
        /// </summary>
        /// <param name="pictureWidth"></param>
        /// <param name="pictureHeight"></param>
        public MapsCollection(int pictureWidth, int pictureHeight)
        {
            _mapStorages = new Dictionary<string, MapWithSetBoatsGeneric<IDrawingObject, AbstractMap>>();
            _pictureWidth = pictureWidth;
            _pictureHeight = pictureHeight;
        }
        /// <summary>
        /// Добавление карты
        /// </summary>
        /// <param name="name">Название карты</param>
        /// <param name="map">Карта</param>
        public void AddMap(string name, AbstractMap map)
        {
            // Добавление карты
            MapWithSetBoatsGeneric<IDrawingObject, AbstractMap> newMap = new(_pictureWidth, _pictureHeight, map);
            _mapStorages.Add(name, newMap);
        }
        /// <summary>
        /// Удаление карты
        /// </summary>
        /// <param name="name">Название карты</param>
        public void DelMap(string name)
        {
            // Удаление карты
            if (!_mapStorages.ContainsKey(name))
            {
                return;
            }
            _mapStorages.Remove(name);
        }
        /// <summary>
        /// Доступ к гавани
        /// </summary>
        /// <param name="ind"></param>
        /// <returns></returns>
        public MapWithSetBoatsGeneric<IDrawingObject, AbstractMap> this[string index]
        {
            get
            {
                // Получение объекта
                if (_mapStorages.ContainsKey(index))
                {
                    return _mapStorages[index];
                }
                return null;
            }
        }
        /// <summary>
        /// Сохранение информации по лодкам в хранилище в файл
        /// </summary>
        /// <param name="filename">Путь и имя файла</param>
        /// <returns></returns>
        public bool SaveData(string filename)
        {
            if (File.Exists(filename))
            {
                File.Delete(filename);
            }
            using (StreamWriter sw = File.CreateText(filename))
            {
                sw.WriteLine($"MapsCollection");
                foreach (var storage in _mapStorages)
                {
                    sw.WriteLine($"{storage.Key}{separatorDict}{storage.Value.GetData(separatorDict, separatorData)}");
                }
            }
            return true;
        }
        /// <summary>
        /// Загрузка нформации по лодкам в гавани из файла
        /// </summary>
        /// <param name="filename"></param>
        /// <returns></returns>
        public bool LoadData(string filename)
        {
            if (!File.Exists(filename))
            {
                return false;
            }
            using (StreamReader sr = File.OpenText(filename))
            {
                string? currentLine = sr.ReadLine();
                if (currentLine == null || !currentLine.Contains("MapsCollection"))
                {
                    //если нет такой записи, то это не те данные
                    return false;
                }
                //очищаем записи
                _mapStorages.Clear();
                currentLine = sr.ReadLine();
                while (currentLine != null)
                {
                    var elem = currentLine.Split(separatorDict);
                    AbstractMap map = null;
                    switch (elem[1])
                    {
                        case "SimpleMap":
                            map = new SimpleMap();
                            break;
                        case "OceanMap":
                            map = new OceanMap();
                            break;
                        case "LineMap":
                            map = new LineMap();
                            break;
                    }
                    _mapStorages.Add(
                        elem[0],
                        new MapWithSetBoatsGeneric<IDrawingObject, AbstractMap>(_pictureWidth, _pictureHeight, map)
                    );
                    _mapStorages[elem[0]].LoadData(
                        elem[2].Split(separatorData, StringSplitOptions.RemoveEmptyEntries));
                    currentLine = sr.ReadLine();
                }
            }
            return true;
        }
    }
}
